
#include "sssd.h"
#include "internals.h"

/*
    Start gameaudio system: open device (usually /dev/dsp), setup device
    parameters, initialize tables, start thread.  Returns -1 on error.
    stopAudio does not need to be called if startAudio does not succeed.
*/

#ifdef __LINUX__
  #include "playthrd16s.c"
  #include "playthrd8s.c"
  #include "playthrd16m.c"
  #include "playthrd8m.c"

  #ifdef _INCLUDE_ESD
  #include <esd.h>
  #endif
  
  #ifdef _INCLUDE_ARTS
  #include <arts.h>
  #endif

#endif

/* Proto rather than function for toollibs */
INLINE_PROTO int Log2(int value);

int sssd_startAudio( struct sssdaudio* ga, const char* device_name )
{
    int format, stereo, speed, fragsize;
    void *thread_rtn;
    int skipconfig=0;

#ifndef __NOPRINTING__
           puts("sssd_startAudio");
#endif

    /* Init variables */
    ga->paused      = 0;
    ga->num_samples = 0;
    ga->num_playing = 0;
    ga->playing_count = 0;
    ga->background_stream = NULL;
    ga->background_stream_bytes = 0;
    ga->threadSignal=0;
    ga->lastERR=0;
    
    /* Allocate arrays */
#ifndef __NOPRINTING__
    printf("sssd: requesting %d total samples\n",ga->max_samples);
#endif
    ga->samples=malloc(sizeof(struct ga_sample)*ga->max_samples);
    if(ga->samples==NULL) {
       ga->lastERR=ERR_SAMPLES_ALLOC;
       return ERR_SAMPLES_ALLOC;
    }
#ifndef __NOPRINTING__
    printf("sssd: requesting %d playing voices\n",ga->max_voices);
#endif
    ga->playing=malloc(sizeof(struct ga_playing)*ga->max_voices);    
    if(ga->playing==NULL) {
       free(ga->samples);
       ga->lastERR=ERR_VOICES_ALLOC;
       return ERR_VOICES_ALLOC;
    }

    if(ga->audioBufferSize==0) {
       ga->audioBufferSize=256*ga->byps;
       if(ga->freq>22000) ga->audioBufferSize<<=1;
       if(ga->freq>44000) ga->audioBufferSize<<=1;         
    }

#ifndef __NOPRINTING__
           printf("sssd: requesting a playbuffer of %d bytes\n",ga->audioBufferSize);
#endif
    
    ga->playBuffer=malloc(ga->audioBufferSize);
    if(ga->playBuffer==NULL) {
       free(ga->samples);
       free(ga->playing);
       ga->lastERR=ERR_PLAYBUFFER_ALLOC;
       return ERR_PLAYBUFFER_ALLOC;
    }
    
#ifndef __NOPRINTING__
           puts("starting device");
#endif
    
    /* Setup audio device */
    if(device_name==DEFAULT_DEVICE) {
       ga->audio_fd = open( DEFAULT_DEVICE_NAME, O_WRONLY, 0 );
    } else
    #ifdef _INCLUDE_ESD
    if(strcmp(device_name,ESD_DEVICE_NAME)==0) {
       /* Setup ESD format parameters */
       
       if(ga->bits==16) format=ESD_BITS16;
       else format=ESD_BITS8;
       
       if(ga->channels==2) format|=ESD_STEREO;
       else format|=ESD_MONO;
       
       format|=ESD_PLAY | ESD_STREAM;
       
       ga->audio_fd = esd_play_stream( format, ga->freq, NULL, NULL);
       skipconfig=1;
       if(ga->audio_fd!=-1) { 
          format=ga->audioBufferSize*ga->numAudioBuffers;
          setsockopt( ga->audio_fd, SOL_SOCKET, SO_SNDBUF, &format, sizeof( format ) );
       }	  

    } else /* end ESD */
    #endif
    { 
       ga->audio_fd = open( device_name, O_WRONLY, 0 );
    }
       
    if(ga->audio_fd== -1) {
    /* Opening default device failed */  
       #ifndef __NOPRINTING__
       perror( device_name );
       #endif
       ga->lastERR=ERR_DEVICE_OPEN;
       return( ERR_DEVICE_OPEN );
    }

    /* Setup device parameters */
    if(ga->bits==16) {
       format = AFMT_S16_LE;
       if(ga->channels==1) thread_rtn= (void*) playthrd16m;
       else thread_rtn=(void *) playthrd16s;       
    } else {
       format = AFMT_U8;
       if(ga->channels==1) thread_rtn= (void*) playthrd8m;
       else thread_rtn=(void *) playthrd8s;
    }
    
    if(skipconfig) goto postOSSconfig;

    stereo   = 0;
    if(ga->channels!=1) stereo = 1;
    speed    = ga->freq;


    /* Program audio device with parameters */

    fragsize=Log2(ga->audioBufferSize)|(ga->numAudioBuffers<<16);
    
    
#ifndef __NOPRINTING__
       printf("Requesting audio buffer block size of %d\n",ga->audioBufferSize);
#endif
       if( ioctl(ga->audio_fd, SNDCTL_DSP_SETFRAGMENT, &fragsize) == -1 )
       {
#ifndef __NOPRINTING__
          perror("SNDCTL_DSP_SETFRAGMENT");  
#endif
          ga->lastERR=ERR_HARDWARE_FRAG;
	  goto err;
       }


    if( ioctl(ga->audio_fd, SNDCTL_DSP_SETFMT, &format) == -1 )
    {
#ifndef __NOPRINTING__
        perror("SNDCTL_DSP_SETFMT");  
#endif
        ga->lastERR=ERR_HARDWARE_BITS;
        goto err;
    }
    if( ioctl(ga->audio_fd, SNDCTL_DSP_STEREO, &stereo) == -1 )
    {
#ifndef __NOPRINTING__
        perror("SNDCTL_DSP_STEREO");  
#endif
        ga->lastERR=ERR_HARDWARE_CHAN;
        goto err;
    }
    if( ioctl(ga->audio_fd, SNDCTL_DSP_SPEED, &speed) == -1 )
    {
#ifndef __NOPRINTING__
        perror("SNDCTL_DSP_SPEED");  
#endif
        ga->lastERR=ERR_HARDWARE_FREQ;
        goto err;
    }

    if( ioctl(ga->audio_fd, SNDCTL_DSP_GETBLKSIZE, &fragsize) == -1 )
    {
#ifndef __NOPRINTING__
        perror("SNDCTL_DSP_GETBLKSIZE");  
#endif
        ga->lastERR=ERR_HARDWARE_PARMS;
        goto err;
    }
    ga->audioBufferSize=fragsize;  
#ifndef __NOPRINTING__
    printf("audio buffer size (bytes)=%d\n",fragsize);
#endif
    if(ga->audioBufferSize<256) printf("WARNING: audio buffer less than 256 bytes\n");

postOSSconfig:  /* place for ESD and aRts to go */

    /* Init mutex */ 
    if( pthread_mutex_init( &ga->buffer_mutex, NULL ) == -1 )
    {
#ifndef __NOPRINTING__
        ga->lastERR=ERR_MUTEX_INIT;
        perror( "pthread_mutex_init()" );
#endif
        goto err;
    }

    /* Start playing thread */
    if( pthread_create( &ga->audio_thread, NULL, (void*) thread_rtn,
                        (void*) ga) == -1 )
    {
#ifndef __NOPRINTING__
        perror( "pthread_create()" );
#endif
        ga->lastERR=ERR_THREAD_START;
        goto err;
    }

    ga->lastERR=0;
    return( 0 );
    
err:

    if(ga->samples!=NULL) free(ga->samples);
    if(ga->playing!=NULL) free(ga->playing);
    close( ga->audio_fd );
    ga->audio_fd = -1;
    return(ga->lastERR);  
}


inline int Log2(int value) {
   int count=0;
   while(!(value&0x1)) {count++;value>>=1;}
return count;
}
